<?php

namespace Drupal\search_api_vragen_ai\Plugin\search_api\backend;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\Backend\BackendPluginBase;
use Drupal\search_api\IndexInterface;
use Drupal\search_api\Item\ItemInterface;
use Drupal\search_api\Plugin\PluginFormTrait;
use Drupal\search_api\Query\QueryInterface;
use Drupal\search_api_vragen_ai\Client;
use Drupal\search_api_vragen_ai\Client\DocumentRepository;
use Http\Discovery\Psr18ClientDiscovery;
use Swis\JsonApi\Client\DocumentFactory;
use Swis\JsonApi\Client\Exceptions\ValidationException;
use Swis\JsonApi\Client\Interfaces\CollectionDocumentInterface;
use Swis\JsonApi\Client\Interfaces\ItemInterface as JsonApiItemInterface;
use Swis\JsonApi\Client\ItemHydrator;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The Vragen.ai search backend.
 *
 * @SearchApiBackend(
 *   id = "vragen_ai",
 *   label = @Translation("Vragen.ai"),
 *   description = @Translation("Vragen.ai semantic search.")
 * )
 */
class VragenAiBackend extends BackendPluginBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * The document repository.
   */
  protected DocumentRepository $documentRepository;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected ConfigFactoryInterface $configFactory;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $instance->withConfigFactory($container->get('config.factory'))
      ->withDocumentRepository(self::buildDefaultDocumentRepository($container->get('config.factory')));

    return $instance;
  }

  /**
   * Set document repository to use.
   *
   * @param \Drupal\search_api_vragen_ai\Client\DocumentRepository $repository
   *   Document repository to use.
   *
   * @return $this
   *   The current instance.
   */
  public function withDocumentRepository(DocumentRepository $repository): self {
    $this->documentRepository = $repository;

    return $this;
  }

  /**
   * Set config factory.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config factory.
   *
   * @return $this
   *   The current instance.
   */
  public function withConfigFactory(ConfigFactoryInterface $configFactory): self {
    $this->configFactory = $configFactory;

    return $this;
  }

  /**
   * Document repository builder.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config factory.
   *
   * @return \Drupal\search_api_vragen_ai\Client\DocumentRepository
   *   Document repository.
   */
  protected static function buildDefaultDocumentRepository(ConfigFactoryInterface $configFactory): DocumentRepository {
    $config = $configFactory->get('search_api_vragen_ai.settings');
    $httpClient = Psr18ClientDiscovery::find();
    $client = Client::create(Client::getTypeMapper(), $httpClient);

    /** @var string $endpoint */
    $endpoint = $config->get('vragen_ai_endpoint') ?? '';

    /** @var string $token */
    $token = $config->get('vragen_ai_token') ?? '';

    $client->setBaseUri($endpoint);
    /** @var \Swis\JsonApi\Client\Client $baseClient */
    $baseClient = $client->getClient();
    $baseClient->setDefaultHeaders(array_merge(
      $baseClient->getDefaultHeaders(), [
        'Authorization' => sprintf('Bearer %s', $token),
      ]
    ));

    return new DocumentRepository(
      $client,
      new DocumentFactory()
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function indexItems(IndexInterface $index, array $items): array {
    $indexed = [];

    foreach ($items as $id => $item) {
      try {
        $this->indexItem($item);
        $indexed[] = $id;
      }
      catch (\Throwable $e) {
        $this->getLogger()->error($e->getMessage() . ' (Document: ' . $item->getId() . ')');
      }
    }

    return $indexed;
  }

  /**
   * Indexes the specified items.
   */
  protected function indexItem(ItemInterface $item): void {
    $content = array_map(function ($field) {
      return implode(' ', array_map(fn ($value) => (string) $value, $field->getValues()));
    }, $item->getFields());

    $content = $content['rendered_item'] ?? implode("\n", $content);
    $content = sprintf('<!doctype html><html><body>%s</body></html>', $content);

    $document = $this->firstOrCreateDocument($item->getId());
    $entity = $item->getOriginalObject();

    if (!$entity) {
      return;
    }

    $publishedDate = date('Y-m-d', (int) $entity->get('created')->getString());

    $document->fill([
      'title' => $item->getDatasource()->getItemLabel($entity) ?? 'Empty',
      'url' => $item->getDatasource()->getItemUrl($entity)?->setAbsolute()->toString(),
      'mime_type' => 'text/html',
      'content' => $content,
      'meta_data' => [
        'entity_type' => $item->getDatasource()->getEntityTypeId(),
        'bundle' => $item->getDatasource()->getItemBundle($entity),
        'date' => $publishedDate,
      ],
    ]);

    if ($document->isNew()) {
      $result = $this->documentRepository->create($document);
    }
    else {
      $result = $this->documentRepository->update($document);
    }

    if (!empty($result->getErrors()->isNotEmpty())) {
      $error = $result->getErrors()->first();
      throw new \Exception('Error when indexing item: ' . $error->getTitle() . ' (' . $error->getStatus() . ': ' . $error->getDetail() . ')');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteItems(IndexInterface $index, array $item_ids): void {
    foreach ($item_ids as $id) {
      $document = $this->firstOrCreateDocument($id);
      if ($document->isNew() || empty($document->getId())) {
        continue;
      }

      $this->documentRepository->delete($document->getId());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteAllIndexItems(IndexInterface $index, $datasource_id = NULL): void {
    // Not supported.
  }

  /**
   * {@inheritdoc}
   */
  public function search(QueryInterface $query): void {
    // Not supported.
  }

  /**
   * {@inheritdoc}
   */
  public function isAvailable(): bool {
    $url = $this->configFactory->get('vragen_ai.settings')->get('vragen_ai_endpoint') ?? '';

    if (empty($url)) {
      $this->getLogger()->error("Vragen AI Endpoint is not set.");
      return FALSE;
    }

    if (filter_var($url, FILTER_VALIDATE_URL) === FALSE) {
      $this->getLogger()->error("Value set as Vragen AI Endpoint is invalid.");
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Get first document or create and return one.
   */
  protected function firstOrCreateDocument(string $id): JsonApiItemInterface {
    $externalReference = mb_strtolower($this->configFactory->get('system.site')->get('name')) . ':' . $id;

    try {
      $result = $this->documentRepository->all([
        'filter' => [
          'external_reference' => $externalReference,
        ],
        'page' => ['size' => 1, 'number' => 1],
      ]);

      if ($result instanceof CollectionDocumentInterface && $item = $result->getData()->first()) {
        /** @var \Swis\JsonApi\Client\Interfaces\ItemInterface $item */
        return $item;
      }
    }
    catch (ValidationException $e) {
      // When response failed to validate, we can safely assume
      // that the document doesn't exist yet.
      $this->getLogger()->notice('Failed to receive existing document: ' . $e->getMessage());
    }

    $itemHydrator = new ItemHydrator(Client::getTypeMapper());
    return $itemHydrator->hydrate(
      Client::getTypeMapper()->getMapping('documents'),
      ['external_reference' => $externalReference]
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __sleep(): array {
    $properties = array_flip(parent::__sleep());
    unset($properties['documentRepository']);
    return array_keys($properties);
  }

  /**
   * {@inheritdoc}
   */
  public function __wakeup(): void {
    parent::__wakeup();

    $this->withDocumentRepository(self::buildDefaultDocumentRepository($this->configFactory));
  }

}
